/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ABSTRACTDISKIO_H
#define ABSTRACTDISKIO_H

#include <QtCore/QObject>
#include <QtCore/QIODevice>
#include <QtCore/QFile>
#include <datasetchain.h>
#include <namedparametertreeitem.h>

class QTime;

/**
 * @brief Disk-IO-things should be in this namespace
 */
namespace NSDiskIO {

	/**
	 * @brief abstract base for DiskIO-things
	 *
	 * opens files for reading/writing and sets the QTextstream _stream accordingly
	 *
	 * @todo enforce every Reader and Writer to give its prefered ending of files
	 * and a description for the selection in FileDialog
	 */
	template<class Stream>
	class AbstractDiskIO
	{
		protected:
			/// normal constructor
			/**
			 * @param openmode the QIODevice::OpenMode for this file
			 */
			AbstractDiskIO( QIODevice::OpenMode openmode )
				: _stream( 0 )
				, _file( 0 )
				, _openmode( openmode )
			{
			}
			/// normal destructor
			/**
			 * calls setFile( "" ) to close the open file
			 */
			virtual ~AbstractDiskIO() {
				setFile( "" );
			}

			/**
			 * @brief the access to the file
			 *
			 * Access the file through this stream. Don't delete this stream!
			 */
			Stream *_stream;

			/**
			 * @brief open a file
			 *
			 * When called with an argument the old opened file is closed and
			 * the new one is opened if it exists.
			 * When the argument is just empty ( "" or QString::Null ) the current file
			 * is closed.
			 *
			 * If opening of the file fails for some reason, it closes the file
			 * and resets the state.
			 *
			 * @param filename The path and name of the file to open.
			 */
			void setFile( QString filename ) {
				qDebug() << "AbstractDiskIO::setFile(" << filename << ")";
				//qDebug() << " _file =" << _file;
				if ( filename == "" && _file != 0 ) {
					delete _stream;
					_stream = 0;
					if ( _file->isOpen() )
						_file->close();
					delete _file;
					_file = 0;
					qDebug() << " old file closed";
				} else {
					if ( _file != 0 )
						setFile( "" );
					if ( _file == 0 && filename != "" ) {
						qDebug() << " opening file" << filename;
						_file = new QFile( filename );
						_file->open( _openmode );
						if ( _file->isOpen() )
							_stream = new Stream( _file );
						else
							setFile( "" );
					}
				}
			}

			/// Return the current filename or ""
			QString filename() const {
				if ( _file )
					return _file->fileName();
				return "";
			}
			/// The current position in the file
			qint64 pos() const {
				if ( _file )
					return _file->pos();
				return 0;
			}
			/// The size of the file
			qint64 fileSize() const {
				if ( _file )
					return _file->size();
				return 0;
			}

			QFile* _file;
		private:
			QIODevice::OpenMode _openmode;
	};

	/**
	 * @brief basic interface for reading file datasets from disk
	 *
	 * This provides a lot of functions to read datasets from disk, the only
	 * thing that is not done in this class is the actual reading...
	 *
	 * You have to implement readSingleDataSet() where you read the data.
	 */
	class AbstractReader : public NSCore::DataSetProducer
	{
		Q_OBJECT
		public:
			/// normal constructor
			AbstractReader( QObject* );
			/// normal destructor
			~AbstractReader();

			/**
			 * @brief The file-endings supported
			 */
			virtual QString filter() const =0;

			/**
			 * @brief the current filename
			 */
			virtual QString filename() const =0;

			/**
			 * @brief whether the end is reached...
			 */
			virtual bool atEnd() const =0;

			/**
			 * @brief Is the current state of type Reading?
			 */
			bool reading() const { return _datasetsToRead != 0; }

			QVariant guiIdentifier( int ) const;

		public slots:
			/// Read one dataset from the file
			void readDataSet();
			/// Read N datasets from the file
			void readNDataSets( int );
			/// Read ten datasets from the file
			void readTenDataSets();
			/// Read 100 datasets from the file
			void readHundredDataSets();
			/**
			 * @brief Read datasets from the file until the end of the file is
			 * reached.
			 */
			void readAllDataSets();

			/**
			 * @brief Go back to start
			 *
			 * Don't collect the money and don't drink any champaign...
			 */
			virtual void reset() =0;

			virtual void setFile( QString filename ) =0;

			void stopReading() { _datasetsToRead = 0; }

		protected:
			/**
			 * @brief the real work
			 *
			 * Implement this! Read one complete NSCore::AcqisitionDataSet from
			 * _stream and enqueue it to the buffer. You don't need to call
			 * signalData() as this will be done by the callers of this
			 * function.
			 *
			 * This function will only be called when a file is open but it
			 * doesn't hurt to check _stream!=0 in the beginning.
			 *
			 * You should emit percent( int ) regularly so the GUI and the user
			 * get informed about how much data i already read.
			 */
			virtual void readSingleDataSet() =0;

			QString debug() const { return "AbstractReader"; }

			NSCore::NamedParameterTreeItemPointer config;

		signals:
			/**
			 * @brief the number of datasets read since the last call to setFile()
			 */
			void readDataSets( int );

		private slots:
			void privateReadPendingDataSets();
			void privateLateInit();

		private:
			QDomNode saveToXML( QDomDocument );
			bool restoreFromXML( QDomNode );

			int _datasetsToRead;

			int _datasetsWritten;

			QTimer* _timer;
	};

	/**
	 * @brief basic interface for writing datasets into files on disk
	 */
	class AbstractWriter : public NSCore::DataSetChainMember {
		Q_OBJECT
		public:
			/// constructor
			AbstractWriter( QObject* );

			/**
			 * @brief Write a single dataset to disk
			 *
			 * You have to implement this function to write a dataset to disk.
			 * There is no need for other things to do, the rest will be done
			 * by customEvent of the AbstractDiskWriter
			 *
			 * @param data the dataset you should write
			 */
			virtual void writeDataSet( NSCore::AcqisitionDataSet& data ) =0;

			/**
			 * @brief The ending of your files
			 *
			 * You have to return the ending of your files without the dot.
			 *
			 * @deprecated
			 */
			virtual QString ending() const __attribute__( ( deprecated ) ) =0;

			virtual QString filter() const { return "All Files (*.*)"; }

			bool dataArrived( NSCore::AcqisitionDataSet& );

		public slots:
			/**
			 * @brief open a file
			 */
			virtual void setFile( QString ) =0;

			/**
			 * @brief resets the counter of the written datasets
			 */
			void resetCounter();

		signals:
			/**
			 * @brief Emits the number of datasets written so far
			 */
			void wroteDataSet( int );

		protected:
			NSCore::NamedParameterTreeItemPointer config;

		private slots:
			void privateLateInit();

		private:
			QString debug() const { return "AbstractDiskWriter"; }

			QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "writer" ); }
			bool restoreFromXML( QDomNode node ) { return node.nodeName() == "writer"; }

			void dataAvailable( NSCore::DataSetProducer* );

			int _number_of_written_datasets;
	};

};

#endif // ABSTRACTDISKIO_H

