/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DISKIOBUFFERS_H
#define DISKIOBUFFERS_H

#include <QtCore/QObject>
#include <QtCore/QIODevice>
#include "abstractdiskio.h"

class QFile;

namespace NSCore {
	class MathTreeRoot;
};

namespace NSDiskIO {

	/**
	 * @brief Read AcqirisLive-Files
	 *
	 * This class reads AcqirisLive files which are always single-channel. So
	 * they aren't of much use except for first testing. :-(
	 */
	class AcqirisLiveReader : public AbstractReader, protected AbstractDiskIO<QTextStream> {
		Q_OBJECT
		public:
			/// constructor
			/**
			 * You have to set the file to open with setFile( QString ).
			 */
			AcqirisLiveReader( QObject* );

		private:
			/// Read one dataset from the file
			void readSingleDataSet();

			double _samplinginterval;
			double _channelfsr;
			double _fullscale;
			double _horizontalposition, _nexthorizontalposition;
			double _lsb;
			QList<double> _data;
	};

	/**
	 * @brief write CSV-files for gnuplot
	 *
	 * This writes CSV files which gives the posibility to load and show the
	 * waveforms in gnuplot.
	 *
	 * Columns are: sample, ch1, ch2,... <br>
	 * The blocks are separated by two newlines, just as gnuplot likes it.
	 */
	class CSVWriter : public AbstractWriter, protected AbstractDiskIO<QTextStream> {
		Q_OBJECT
		public:
			CSVWriter( QObject* =0 );

			QString ending() const { return "csv"; }

			void writeDataSet( NSCore::AcqisitionDataSet& data );

		public slots:
			void setFile( QString filename ) {
				qDebug() << "CombinedCSV::setFile(" << filename << ")";
				AbstractDiskIO<QTextStream>::setFile( filename );
			}

		private:
			NSCore::MathTreeRoot *_timealign;
			NSCore::MathTreeRoot *_timeinterval;
	};

	/**
	 * @brief EPOS own format
	 */
	class EPOSWriter : public AbstractWriter, protected AbstractDiskIO<QTextStream> {
		public:
			EPOSWriter( QObject* );

			void writeDataSet( NSCore::AcqisitionDataSet& data );
	};

	/**
	 * @brief Load EPOS own format
	 */
	class EPOSReader : public AbstractReader, protected AbstractDiskIO<QTextStream> {
		public:
			EPOSReader( QObject* );
			QString filter() const { return "EPOS file (*.eposfile)"; }

			void readSingleDataSet();

			void reset() {
				_stream->reset();
			}
			void setFile( QString file );
			QString filename() const;
			bool atEnd() const {
				return _stream->atEnd();
			}
	};

	/**
	 * @brief Combined CSV-Files for Origin/...
	 */
	class CombinedCSV : public AbstractWriter, protected AbstractDiskIO<QTextStream> {
		Q_OBJECT
		public:
			CombinedCSV( QObject* =0 );

			QString ending() const { return "csv"; }

			void writeDataSet( NSCore::AcqisitionDataSet& data );

		public slots:
			void setFile( QString filename ) {
				qDebug() << "CombinedCSV::setFile(" << filename << ")";
				AbstractDiskIO<QTextStream>::setFile( filename );
			}

		private:
			QMap< int,QList<double> > _save;
			int _combine_counter;
	};

};

#endif // DISKIOBUFFERS_H

