/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "evaluationpath.h"
#include "evaluationpath.moc"

#include "evaluationpath_items.h"
#include "evaluationpath_items.moc"
#include "plugins_interface.h"

#include "mathexpression.h"
#include "filterexpression.h"

#include <cmath>

#include <QtCore/QDebug>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <QtCore/QTimer>
#include <QtXml/QDomDocument>

namespace NSEvaluation {

PluginItem* pluginFromXML( QDomElement elem, NSCore::QTreeItem* parent ) {
	QString plugin = elem.attribute( "type" );
	QString plugintype = PluginFactory::the()->type( plugin );
	//qDebug() << "pluginFromXML() type =" << plugin << "plugintype =" << plugintype;
	PluginItem* ret = 0;
	if ( plugintype == "Analyzer" )
		ret = new AnalyzerItem( plugin, parent );
	if ( plugintype == "Modifier" )
		ret = new ModifierItem( plugin, parent );
	if ( plugintype == "Filter" )
		ret = new FilterItem( plugin, parent );
	if ( ret )
		ret->fromXml( elem );
	return ret;
}

EvaluationPath::EvaluationPath( QObject* p, NSCore::QTreeItem* t )
	: NSCore::DataSetChainMember( t )
	, _config( new NSCore::NamedParameterTreeItem( "config", QVariant::Invalid, this ) )
	, _timer( new QTimer( this ) )
{
	setObjectName( "EvaluationPath" );
	qDebug() << "EvaluationPath::EvaluationPath(" << p << ")";

	_config->addParameter( "processed", QVariant( 0 ) );
	_config->getParameter( "processed" )->setMin( 0 );
	_config->getParameter( "processed" )->setMax( int( 1E10 ) );
	new FilterExpression( _config );
	//_config->addParameter( "through", QVariant( 0 ) );
	//_config->getParameter( "through" )->setMin( 0 );
	//_config->getParameter( "through" )->setMax( int( 1E10 ) );
	_config->getParameter( "processed" )->setGuiHint( NSCore::NamedParameterTreeItem::Counter );
	//_config->getParameter( "through" )->setGuiHint( NSCore::NamedParameterTreeItem::Counter );

	qDebug() << "EvaluationPath::EvaluationPath() finished";

	_timer->setInterval( 1000 );
	//_timer->setSingleShot( true );
	connect( _timer, SIGNAL( timeout() ), this, SLOT( checkBuffers() ) );
	_timer->start();
}
EvaluationPath::~EvaluationPath() {
	qDebug() << "EvaluationPath::~EvaluationPath()";
	_timer->stop();
}

bool EvaluationPath::dataArrived( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "EvaluationPath::dataArrived(" << data << ")";
	_config->getParameter( "processed" )->setValue( _config->getParameter( "processed" )->toInt() + 1 );

	foreach( PluginItem* plugin, childobjects<PluginItem>() )
		plugin->process( data );

	if ( ! isnan( _config->child<FilterExpression>( "Filter" )->value( data ) ) )
		return true;

	return false;
}

void EvaluationPath::checkBuffers() {
	if ( thread() != QThread::currentThread() )
		qDebug() << "EvaluationPath::checkBuffers() my thread:" << thread() << " current thread:" << QThread::currentThread();

	//_config->child<FilterExpression>( "Filter" )->workOnCache();

	while ( ! _pluginsToDelete.isEmpty() )
		delete _pluginsToDelete.takeFirst();

	while ( ! _xmlToParse.isEmpty() ) {
		//qDebug() << "EvaluationPath::checkBuffers() [" << this << "] parsing xml";
		xml_cache_t xml = _xmlToParse.takeFirst();
		PluginItem* tmp = pluginFromXML( xml.xml, xml.parent );
		//qDebug() << "EvaluationPath::checkBuffers() [" << this << "] plugin returned is" << tmp;
		connect(
			tmp, SIGNAL( sExecAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
			this, SLOT( execAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
			Qt::DirectConnection );
		connect(
			tmp, SIGNAL( sExecRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
			this, SLOT( execRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
			Qt::DirectConnection );
		//qDebug() << "EvaluationPath::checkBuffers() [" << this << "] done, isEmpty?" << _xmlToParse.isEmpty();
	}

	if ( ! _timer->isActive() )
		_timer->start();
}

void EvaluationPath::execAdd( NSCore::QTreeItem* parent, NSCore::GuiServer_Interface* gui ) {
	QList<NSCore::GuiServer_Interface::ListItem> items;
	foreach( QString plugin, PluginFactory::the()->plugins() ) {
		NSCore::GuiServer_Interface::ListItem item;
		item.title = plugin;
		item.help = PluginFactory::the()->help( plugin );
		item.categories += PluginFactory::the()->categories( plugin );
		if ( ! item.categories.contains( PluginFactory::the()->type( plugin ) ) )
			item.categories += PluginFactory::the()->type( plugin );
		items.append( item );
	}
	QString plugin = gui->getItem( "Select the new plugin", "Select the new plugin you want to add to the tree:", items );
	qDebug() << " returned" << plugin;
	if ( ! plugin.isEmpty() ) {
		QString type = PluginFactory::the()->type( plugin );
		QDomDocument doc( "bla" );
		doc.setContent( QString( "<%1 type=\"%2\" />" ).arg( type ).arg( plugin ) );
		xml_cache_t tmp;
		tmp.xml = doc.documentElement();
		tmp.parent = parent;
		_xmlToParse.append( tmp );
	}
}
void EvaluationPath::execRemove( NSCore::QTreeItem* p, NSCore::GuiServer_Interface* gui ) {
	PluginItem* plugin = dynamic_cast<PluginItem*>( p );
	if ( !plugin )
		return;
	bool yes = gui->messageYesNo( "Are you sure?", "Do you really want to remove this plugin?" );
	if ( yes )
		_pluginsToDelete.append( plugin );
}

void EvaluationPath::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	if ( action == "Add" )
		execAdd( this, gui );
	else
		NSCore::QTreeItem::guiExecAction( action, gui );
}
QVariant EvaluationPath::guiIdentifier( int role ) const {
	switch ( role ) {
		case Qt::DecorationRole:
			return ":/icons/cr48-action-evaluate_pulses.png";
			break;
		default:
			return QTreeItem::guiIdentifier( role );
	}
}

bool EvaluationPath::createFromXML( QDomNode node ) {
	//qDebug() << "EvaluationPath::createFromXML()";
	foreach( QObject* obj, childobjects() ) {
		PluginItem* tmp = dynamic_cast<PluginItem*>( obj );
		if ( tmp )
			_pluginsToDelete.append( tmp );
	}
	QDomElement elem = node.firstChildElement();
	while ( !elem.isNull() ) {
		if ( elem.tagName() != "parameter" ) {
			xml_cache_t tmp;
			tmp.xml = elem;
			tmp.parent = this;
			_xmlToParse.append( tmp );
		}
		if ( elem.tagName() == "parameter" ) {
			_config->fromXml( elem );
		}
		elem = elem.nextSiblingElement();
	}
	return false;
}


bool PluginItem::createFromXML( QDomNode node ) {
	//qDebug() << "PluginItem::createFromXML()";
	QDomElement elem = node.firstChildElement();
	QStringList tags;
	tags << "analyzer" << "modifier" << "filter";
	while ( !elem.isNull() ) {
		if ( tags.contains( elem.tagName() ) ) {
			PluginItem* tmp = pluginFromXML( elem, this );
			connect(
				tmp, SIGNAL( sExecAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
				this, SIGNAL( sExecAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
				Qt::DirectConnection );
			connect(
				tmp, SIGNAL( sExecRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
				this, SIGNAL( sExecRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* ) ),
				Qt::DirectConnection );
		} else
			config()->fromXml( elem );
		elem = elem.nextSiblingElement();
	}
	return false;
}

};

